@@ -56,8 +56,6 @@ static const char *type2string(enum connman_element_type type)
 		return "network";
 	case CONNMAN_ELEMENT_TYPE_SERVICE:
 		return "service";
-	case CONNMAN_ELEMENT_TYPE_PPP:
-		return "ppp";
 	case CONNMAN_ELEMENT_TYPE_IPV4:
 		return "ipv4";
 	case CONNMAN_ELEMENT_TYPE_IPV6:
@@ -253,18 +251,13 @@ struct connman_service *__connman_element_get_service(struct connman_element *el
 	case CONNMAN_DEVICE_TYPE_UNKNOWN:
 	case CONNMAN_DEVICE_TYPE_VENDOR:
 	case CONNMAN_DEVICE_TYPE_GPS:
-	case CONNMAN_DEVICE_TYPE_NOZOMI:
-	case CONNMAN_DEVICE_TYPE_HUAWEI:
-	case CONNMAN_DEVICE_TYPE_NOVATEL:
 		break;
 	case CONNMAN_DEVICE_TYPE_ETHERNET:
-		service = __connman_service_lookup_from_device(device);
-		break;
 	case CONNMAN_DEVICE_TYPE_WIFI:
 	case CONNMAN_DEVICE_TYPE_WIMAX:
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
-	case CONNMAN_DEVICE_TYPE_MBM:
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
+	case CONNMAN_DEVICE_TYPE_MBM:
 	case CONNMAN_DEVICE_TYPE_HSO:
 		network = __connman_element_get_network(element);
 		if (network == NULL)
@@ -361,9 +354,11 @@ static gboolean request_scan(GNode *node, gpointer user_data)
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_SYSTEM:
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+	case CONNMAN_SERVICE_TYPE_VPN:
 		return FALSE;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
@@ -404,6 +399,8 @@ static gboolean enable_technology(GNode *node, gpointer user_data)
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_SYSTEM:
+	case CONNMAN_SERVICE_TYPE_VPN:
 		return FALSE;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIFI:
@@ -447,6 +444,8 @@ static gboolean disable_technology(GNode *node, gpointer user_data)
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_SYSTEM:
+	case CONNMAN_SERVICE_TYPE_VPN:
 		return FALSE;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIFI:
@@ -683,6 +682,7 @@ void connman_element_unref(struct connman_element *element)
 		g_free(element->ipv4.network);
 		g_free(element->ipv4.broadcast);
 		g_free(element->ipv4.nameserver);
+		g_free(element->ipv4.timeserver);
 		g_free(element->devname);
 		g_free(element->path);
 		g_free(element->name);
@@ -868,6 +868,14 @@ int connman_element_get_value(struct connman_element *element,
 		*((char **) value) = element->ipv4.nameserver;
 		__connman_element_unlock(element);
 		break;
+	case CONNMAN_PROPERTY_ID_IPV4_TIMESERVER:
+		if (element->ipv4.timeserver == NULL)
+			return connman_element_get_value(element->parent,
+								id, value);
+		__connman_element_lock(element);
+		*((char **) value) = element->ipv4.timeserver;
+		__connman_element_unlock(element);
+		break;
 	default:
 		return -EINVAL;
 	}
@@ -1091,6 +1099,7 @@ int __connman_element_append_ipv4(struct connman_element *element,
 	const char *method = NULL;
 	const char *address = NULL, *netmask = NULL, *gateway = NULL;
 	const char *broadcast = NULL, *nameserver = NULL;
+	const char *timeserver = NULL;
 
 	connman_element_get_value(element,
 				CONNMAN_PROPERTY_ID_IPV4_METHOD, &method);
@@ -1105,31 +1114,37 @@ int __connman_element_append_ipv4(struct connman_element *element,
 			CONNMAN_PROPERTY_ID_IPV4_BROADCAST, &broadcast);
 	connman_element_get_value(element,
 			CONNMAN_PROPERTY_ID_IPV4_NAMESERVER, &nameserver);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV4_TIMESERVER, &timeserver);
 
 	if (method != NULL)
-		connman_dbus_dict_append_variant(dict, "IPv4.Method",
+		connman_dbus_dict_append_basic(dict, "IPv4.Method",
 						DBUS_TYPE_STRING, &method);
 
 	if (address != NULL)
-		connman_dbus_dict_append_variant(dict, "IPv4.Address",
+		connman_dbus_dict_append_basic(dict, "IPv4.Address",
 						DBUS_TYPE_STRING, &address);
 
 	if (netmask != NULL)
-		connman_dbus_dict_append_variant(dict, "IPv4.Netmask",
+		connman_dbus_dict_append_basic(dict, "IPv4.Netmask",
 						DBUS_TYPE_STRING, &netmask);
 
 	if (gateway != NULL)
-		connman_dbus_dict_append_variant(dict, "IPv4.Gateway",
+		connman_dbus_dict_append_basic(dict, "IPv4.Gateway",
 						DBUS_TYPE_STRING, &gateway);
 
 	if (broadcast != NULL)
-		connman_dbus_dict_append_variant(dict, "IPv4.Broadcast",
+		connman_dbus_dict_append_basic(dict, "IPv4.Broadcast",
 						DBUS_TYPE_STRING, &broadcast);
 
 	if (nameserver != NULL)
-		connman_dbus_dict_append_variant(dict, "IPv4.Nameserver",
+		connman_dbus_dict_append_basic(dict, "IPv4.Nameserver",
 						DBUS_TYPE_STRING, &nameserver);
 
+	if (timeserver != NULL)
+		connman_dbus_dict_append_basic(dict, "IPv4.Timeserver",
+						DBUS_TYPE_STRING, &timeserver);
+
 	return 0;
 }
 
@@ -1217,50 +1232,40 @@ int __connman_element_set_ipv4(struct connman_element *element,
 		g_free(element->ipv4.nameserver);
 		element->ipv4.nameserver = g_strdup(nameserver);
 
+		connman_element_update(element);
+	} else if (g_str_equal(name, "IPv4.Timeserver") == TRUE) {
+		const char *timeserver;
+
+		if (type != DBUS_TYPE_STRING)
+			return -EINVAL;
+
+		dbus_message_iter_get_basic(value, &timeserver);
+
+		g_free(element->ipv4.timeserver);
+		element->ipv4.nameserver = g_strdup(timeserver);
+
 		connman_element_update(element);
 	}
 
 	return 0;
 }
 
-static void append_state(DBusMessageIter *entry, const char *state)
-{
-	DBusMessageIter value;
-	const char *key = "State";
-
-	dbus_message_iter_append_basic(entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_STRING_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &state);
-	dbus_message_iter_close_container(entry, &value);
-}
-
 static void emit_state_change(DBusConnection *conn, const char *state)
 {
 	DBusMessage *signal;
-	DBusMessageIter entry;
+	DBusMessageIter iter;
 
-	DBG("conn %p", conn);
-
-	signal = dbus_message_new_signal(CONNMAN_MANAGER_PATH,
-				CONNMAN_MANAGER_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	append_state(&entry, state);
-
-	g_dbus_send_message(conn, signal);
+	connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
+				CONNMAN_MANAGER_INTERFACE, "State",
+						DBUS_TYPE_STRING, &state);
 
 	signal = dbus_message_new_signal(CONNMAN_MANAGER_PATH,
 				CONNMAN_MANAGER_INTERFACE, "StateChanged");
 	if (signal == NULL)
 		return;
 
-	dbus_message_iter_init_append(signal, &entry);
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &state);
+	dbus_message_iter_init_append(signal, &iter);
+	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &state);
 
 	g_dbus_send_message(conn, signal);
 }
@@ -1586,6 +1591,7 @@ int __connman_element_init(const char *device, const char *nodevice)
 
 	__connman_notifier_init();
 	__connman_service_init();
+	__connman_provider_init();
 	__connman_network_init();
 	__connman_device_init();
 
@@ -1625,6 +1631,7 @@ void __connman_element_start(void)
 
 	__connman_connection_init();
 	__connman_ipv4_init();
+	__connman_dhcp_init();
 
 	if (__connman_rfkill_init() < 0)
 		__connman_udev_enable_rfkill_processing();
@@ -1636,7 +1643,9 @@ void __connman_element_stop(void)
 
 	__connman_rfkill_cleanup();
 
+	__connman_dhcp_cleanup();
 	__connman_ipv4_cleanup();
+	__connman_provider_cleanup();
 	__connman_connection_cleanup();
 }
 
