@@ -45,6 +45,10 @@ struct connman_ipconfig {
 
 	enum connman_ipconfig_method method;
 	struct connman_ipaddress *address;
+	struct connman_ipaddress *system;
+
+	char *eth;
+	uint16_t mtu;
 };
 
 struct connman_ipdevice {
@@ -87,6 +91,54 @@ void connman_ipaddress_free(struct connman_ipaddress *ipaddress)
 	g_free(ipaddress);
 }
 
+static unsigned char netmask2prefixlen(const char *netmask)
+{
+	unsigned char bits = 0;
+	in_addr_t mask = inet_network(netmask);
+	in_addr_t host = ~mask;
+
+	/* a valid netmask must be 2^n - 1 */
+	if ((host & (host + 1)) != 0)
+		return -1;
+
+	for (; mask; mask <<= 1)
+		++bits;
+
+	return bits;
+}
+
+void connman_ipaddress_set(struct connman_ipaddress *ipaddress,
+				const char *address, const char *netmask)
+{
+	if (ipaddress == NULL)
+		return;
+
+	if (netmask != NULL)
+		ipaddress->prefixlen = netmask2prefixlen(netmask);
+	else
+		ipaddress->prefixlen = 32;
+
+	g_free(ipaddress->local);
+	ipaddress->local = g_strdup(address);
+}
+
+void connman_ipaddress_clear(struct connman_ipaddress *ipaddress)
+{
+	if (ipaddress == NULL)
+		return;
+
+	ipaddress->prefixlen = 0;
+
+	g_free(ipaddress->local);
+	ipaddress->local = NULL;
+
+	g_free(ipaddress->peer);
+	ipaddress->peer = NULL;
+
+	g_free(ipaddress->broadcast);
+	ipaddress->broadcast = NULL;
+}
+
 void connman_ipaddress_copy(struct connman_ipaddress *ipaddress,
 					struct connman_ipaddress *source)
 {
@@ -235,8 +287,9 @@ static void __connman_ipconfig_lower_up(struct connman_ipdevice *ipdevice)
 
 	switch (ipdevice->config->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
-	case CONNMAN_IPCONFIG_METHOD_IGNORE:
-	case CONNMAN_IPCONFIG_METHOD_STATIC:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		return;
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 		break;
@@ -285,7 +338,8 @@ static void __connman_ipconfig_lower_down(struct connman_ipdevice *ipdevice)
 }
 
 void __connman_ipconfig_newlink(int index, unsigned short type,
-							unsigned int flags)
+				unsigned int flags, const char *address,
+							unsigned short mtu)
 {
 	struct connman_ipdevice *ipdevice;
 	GList *list;
@@ -327,6 +381,12 @@ void __connman_ipconfig_newlink(int index, unsigned short type,
 						index, type, type2str(type));
 
 update:
+	if (ipdevice->config != NULL) {
+		g_free(ipdevice->config->eth);
+		ipdevice->config->eth = g_strdup(address);
+		ipdevice->config->mtu = mtu;
+	}
+
 	if (flags == ipdevice->flags)
 		return;
 
@@ -346,8 +406,6 @@ update:
 			lower_down = TRUE;
 	}
 
-	connman_inet_clear_address(index);
-
 	ipdevice->flags = flags;
 
 	str = g_string_new(NULL);
@@ -377,6 +435,9 @@ update:
 		if (index != ipconfig->index)
 			continue;
 
+		if (ipconfig->ops == NULL)
+			continue;
+
 		if (up == TRUE && ipconfig->ops->up)
 			ipconfig->ops->up(ipconfig);
 		if (lower_up == TRUE && ipconfig->ops->lower_up)
@@ -414,6 +475,9 @@ void __connman_ipconfig_dellink(int index)
 
 		ipconfig->index = -1;
 
+		if (ipconfig->ops == NULL)
+			continue;
+
 		if (ipconfig->ops->lower_down)
 			ipconfig->ops->lower_down(ipconfig);
 		if (ipconfig->ops->down)
@@ -451,6 +515,9 @@ void __connman_ipconfig_newaddr(int index, const char *label,
 	connman_info("%s {add} address %s/%u label %s", ipdevice->ifname,
 						address, prefixlen, label);
 
+	if (ipdevice->config != NULL)
+		connman_ipaddress_copy(ipdevice->config->system, ipaddress);
+
 	if ((ipdevice->flags & (IFF_RUNNING | IFF_LOWER_UP)) != (IFF_RUNNING | IFF_LOWER_UP))
 		return;
 
@@ -464,6 +531,9 @@ void __connman_ipconfig_newaddr(int index, const char *label,
 		if (index != ipconfig->index)
 			continue;
 
+		if (ipconfig->ops == NULL)
+			continue;
+
 		if (ipconfig->ops->ip_bound)
 			ipconfig->ops->ip_bound(ipconfig);
 	}
@@ -507,6 +577,9 @@ void __connman_ipconfig_deladdr(int index, const char *label,
 		if (index != ipconfig->index)
 			continue;
 
+		if (ipconfig->ops == NULL)
+			continue;
+
 		if (ipconfig->ops->ip_release)
 			ipconfig->ops->ip_release(ipconfig);
 	}
@@ -632,6 +705,8 @@ struct connman_ipconfig *connman_ipconfig_create(int index)
 		return NULL;
 	}
 
+	ipconfig->system = connman_ipaddress_alloc();
+
 	DBG("ipconfig %p", ipconfig);
 
 	return ipconfig;
@@ -685,6 +760,8 @@ struct connman_ipconfig *connman_ipconfig_ref(struct connman_ipconfig *ipconfig)
 void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
 {
 	if (g_atomic_int_dec_and_test(&ipconfig->refcount) == TRUE) {
+		__connman_ipconfig_disable(ipconfig);
+
 		connman_ipconfig_set_ops(ipconfig, NULL);
 
 		if (ipconfig->origin != NULL) {
@@ -692,7 +769,9 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
 			ipconfig->origin = NULL;
 		}
 
+		connman_ipaddress_free(ipconfig->system);
 		connman_ipaddress_free(ipconfig->address);
+		g_free(ipconfig->eth);
 		g_free(ipconfig);
 	}
 }
@@ -765,13 +844,7 @@ const char *connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
 void connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
 				const struct connman_ipconfig_ops *ops)
 {
-	if (ipconfig->ops != NULL)
-		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
-
 	ipconfig->ops = ops;
-
-	if (ops != NULL)
-		ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 }
 
 /**
@@ -822,11 +895,18 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 	if (ipdevice == NULL)
 		return -ENXIO;
 
-	if (ipdevice->config != NULL)
+	if (ipdevice->config != NULL) {
+		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
+
+		connman_ipaddress_clear(ipdevice->config->system);
+
 		connman_ipconfig_unref(ipdevice->config);
+	}
 
 	ipdevice->config = connman_ipconfig_ref(ipconfig);
 
+	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
+
 	return 0;
 }
 
@@ -847,6 +927,10 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 	if (ipdevice->config == NULL || ipdevice->config != ipconfig)
 		return -EINVAL;
 
+	ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
+
+	connman_ipaddress_clear(ipdevice->config->system);
+
 	connman_ipconfig_unref(ipdevice->config);
 	ipdevice->config = NULL;
 
@@ -858,10 +942,12 @@ const char *__connman_ipconfig_method2string(enum connman_ipconfig_method method
 	switch (method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 		break;
-	case CONNMAN_IPCONFIG_METHOD_IGNORE:
-		return "ignore";
-	case CONNMAN_IPCONFIG_METHOD_STATIC:
-		return "static";
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+		return "off";
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+		return "fixed";
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		return "manual";
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 		return "dhcp";
 	}
@@ -871,35 +957,20 @@ const char *__connman_ipconfig_method2string(enum connman_ipconfig_method method
 
 enum connman_ipconfig_method __connman_ipconfig_string2method(const char *method)
 {
-	if (g_strcmp0(method, "ignore") == 0)
-		return CONNMAN_IPCONFIG_METHOD_IGNORE;
-	else if (g_strcmp0(method, "static") == 0)
-		return CONNMAN_IPCONFIG_METHOD_STATIC;
+	if (g_strcmp0(method, "off") == 0)
+		return CONNMAN_IPCONFIG_METHOD_OFF;
+	else if (g_strcmp0(method, "fixed") == 0)
+		return CONNMAN_IPCONFIG_METHOD_FIXED;
+	else if (g_strcmp0(method, "manual") == 0)
+		return CONNMAN_IPCONFIG_METHOD_MANUAL;
 	else if (g_strcmp0(method, "dhcp") == 0)
 		return CONNMAN_IPCONFIG_METHOD_DHCP;
 	else
 		return CONNMAN_IPCONFIG_METHOD_UNKNOWN;
 }
 
-static void append_variant(DBusMessageIter *iter, const char *prefix,
-					const char *key, int type, void *val)
-{
-	char *str;
-
-	if (prefix == NULL) {
-		connman_dbus_dict_append_variant(iter, key, type, val);
-		return;
-	}
-
-	str = g_strdup_printf("%s%s", prefix, key);
-	if (str != NULL)
-		connman_dbus_dict_append_variant(iter, str, type, val);
-
-	g_free(str);
-}
-
 void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
-				DBusMessageIter *iter, const char *prefix)
+							DBusMessageIter *iter)
 {
 	const char *str;
 
@@ -907,28 +978,162 @@ void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 	if (str == NULL)
 		return;
 
-	append_variant(iter, prefix, "Method", DBUS_TYPE_STRING, &str);
+	connman_dbus_dict_append_basic(iter, "Method", DBUS_TYPE_STRING, &str);
+
+	if (ipconfig->system == NULL)
+		return;
+
+	if (ipconfig->system->local != NULL) {
+		in_addr_t addr;
+		struct in_addr netmask;
+		char *mask;
+
+		connman_dbus_dict_append_basic(iter, "Address",
+				DBUS_TYPE_STRING, &ipconfig->system->local);
+
+		addr = 0xffffffff << (32 - ipconfig->system->prefixlen);
+		netmask.s_addr = htonl(addr);
+		mask = inet_ntoa(netmask);
+		connman_dbus_dict_append_basic(iter, "Netmask",
+						DBUS_TYPE_STRING, &mask);
+	}
 }
 
-int __connman_ipconfig_set_ipv4(struct connman_ipconfig *ipconfig,
-				const char *key, DBusMessageIter *value)
+void __connman_ipconfig_append_ipv4config(struct connman_ipconfig *ipconfig,
+							DBusMessageIter *iter)
 {
-	int type = dbus_message_iter_get_arg_type(value);
+	const char *str;
 
-	DBG("ipconfig %p key %s type %d", ipconfig, key, type);
+	str = __connman_ipconfig_method2string(ipconfig->method);
+	if (str == NULL)
+		return;
 
-	if (g_strcmp0(key, "Method") == 0) {
-		const char *method;
+	connman_dbus_dict_append_basic(iter, "Method", DBUS_TYPE_STRING, &str);
 
-		if (type != DBUS_TYPE_STRING)
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+		return;
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		break;
+	}
+
+	if (ipconfig->address == NULL)
+		return;
+
+	if (ipconfig->address->local != NULL) {
+		in_addr_t addr;
+		struct in_addr netmask;
+		char *mask;
+
+		connman_dbus_dict_append_basic(iter, "Address",
+				DBUS_TYPE_STRING, &ipconfig->address->local);
+
+		addr = 0xffffffff << (32 - ipconfig->address->prefixlen);
+		netmask.s_addr = htonl(addr);
+		mask = inet_ntoa(netmask);
+		connman_dbus_dict_append_basic(iter, "Netmask",
+						DBUS_TYPE_STRING, &mask);
+	}
+}
+
+int __connman_ipconfig_set_ipv4config(struct connman_ipconfig *ipconfig,
+							DBusMessageIter *array)
+{
+	enum connman_ipconfig_method method = CONNMAN_IPCONFIG_METHOD_UNKNOWN;
+	const char *address = NULL, *netmask = NULL;
+	DBusMessageIter dict;
+
+	DBG("ipconfig %p", ipconfig);
+
+	if (dbus_message_iter_get_arg_type(array) != DBUS_TYPE_ARRAY)
+		return -EINVAL;
+
+	dbus_message_iter_recurse(array, &dict);
+
+	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter entry;
+		const char *key;
+		int type;
+
+		dbus_message_iter_recurse(&dict, &entry);
+
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_STRING)
 			return -EINVAL;
 
-		dbus_message_iter_get_basic(value, &method);
+		dbus_message_iter_get_basic(&entry, &key);
+		dbus_message_iter_next(&entry);
 
-		ipconfig->method = __connman_ipconfig_string2method(method);
-	} else
+		type = dbus_message_iter_get_arg_type(&entry);
+
+		if (g_str_equal(key, "Method") == TRUE) {
+			const char *str;
+
+			if (type != DBUS_TYPE_STRING)
+				return -EINVAL;
+
+			dbus_message_iter_get_basic(&entry, &str);
+			method = __connman_ipconfig_string2method(str);
+		} else if (g_str_equal(key, "Address") == TRUE) {
+			if (type != DBUS_TYPE_STRING)
+				return -EINVAL;
+
+			dbus_message_iter_get_basic(&entry, &address);
+		} else if (g_str_equal(key, "Netmask") == TRUE) {
+			if (type != DBUS_TYPE_STRING)
+				return -EINVAL;
+
+			dbus_message_iter_get_basic(&entry, &netmask);
+		}
+
+		dbus_message_iter_next(&dict);
+	}
+
+	DBG("method %d address %s netmask %s", method, address, netmask);
+
+	switch (method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
 		return -EINVAL;
 
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		if (address == NULL)
+			return -EINVAL;
+
+		ipconfig->method = method;
+		connman_ipaddress_set(ipconfig->address, address, netmask);
+		break;
+
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+		if (ipconfig->method == method)
+			return 0;
+
+		ipconfig->method = method;
+		break;
+	}
+
+	return 0;
+}
+
+int __connman_ipconfig_append_ethernet(struct connman_ipconfig *ipconfig,
+							DBusMessageIter *iter)
+{
+	const char *method = "auto";
+
+	connman_dbus_dict_append_basic(iter, "Method",
+						DBUS_TYPE_STRING, &method);
+
+	if (ipconfig->eth != NULL)
+		connman_dbus_dict_append_basic(iter, "Address",
+					DBUS_TYPE_STRING, &ipconfig->eth);
+
+	if (ipconfig->mtu > 0)
+		connman_dbus_dict_append_basic(iter, "MTU",
+					DBUS_TYPE_UINT16, &ipconfig->mtu);
+
 	return 0;
 }
 
